var Config, applyColor, createHash, createId, crypto, fs, getWord, home, listDir, p, parseText, pushWords, readline, separatorCharsRe, signType, statePrefix, stateRe, tagPrefix, tagRe, times;
require("colors");
crypto = require("crypto");
signType = "md5";
fs = require("fs");
readline = require("readline");
p = require("path");
home = process.env.HOME + "/";
exports.statePrefix = statePrefix = "@";
exports.tagPrefix = tagPrefix = "+";
stateRe = new RegExp("^@(\\S+)");
tagRe = new RegExp("^\\+(\\S+)");
separatorCharsRe = new RegExp("([^:,.+\\-!@#&=]+)");
/*
Sign string with username and email

@param {String} str Source string
@param {Object} config Config object
@return {String} hash Hex digest hash
@api public
*/
exports.sign = function(str, config) {
  return createHash(str + config.get("user") + config.get("email"), config.get("secret"));
};
/*
Create hash from string

@param {String} str Source string
@return {String} hash Hex digest hash
@api public
*/
exports.createHash = createHash = function(str, secret) {
  if (secret == null) {
    secret = "change me";
  }
  return crypto.createHmac(signType, secret).update(str).digest("hex");
};
/*
Create unique ID for string

@param {String} str Source string
@return {String} hash Unique hex digest hash
@api public
*/
exports.createId = createId = function(str, config) {
  var secret;
  if (config == null) {
    config = null;
  }
  secret = config ? config.get("secret") : "change me";
  return createHash(str + new Date(), secret);
};
/*
Class for set or set config flags
*/
Config = (function() {
  /*
  Load config. Loads config from ~/.strack
  If config file not exists, create it.
  Try to gather user name and email from ~/.gitconfig, if
  file not exists, get user from process.env.USER

  @return {Object} config Config object, that have "user" and "email" keys.
  @api public
  */  function Config() {
    var conf, gitconfig, line, _i, _len, _ref;
    try {
      this.configFile = home + ".strackrc";
      fs.statSync(this.configFile);
      this.config = JSON.parse(fs.readFileSync(this.configFile));
      this._writeDefaults();
    } catch (err) {
      if ('ENOENT' === err.code) {
        this.config = {};
        try {
          gitconfig = home + ".gitconfig";
          fs.statSync(gitconfig);
          conf = fs.readFileSync(gitconfig);
          _ref = conf.toString().split("\n");
          for (_i = 0, _len = _ref.length; _i < _len; _i++) {
            line = _ref[_i];
            line = line.trim();
            if (0 === line.indexOf("name")) {
              this.config.user = line.split("=")[1].trim();
            } else if (0 === line.indexOf("email")) {
              this.config.email = line.split("=")[1].trim();
            }
          }
        } catch (err2) {
          this._writeDefaults();
        }
        fs.writeFileSync(this.configFile, JSON.stringify(this.config));
      } else {
        throw err;
      }
    }
  }
  /*
  Write defaults to config in they not presents

  @api private
  */
  Config.prototype._writeDefaults = function() {
    var _base, _base10, _base11, _base12, _base13, _base2, _base3, _base4, _base5, _base6, _base7, _base8, _base9;
    (_base = this.config).user || (_base.user = process.env.USER);
    if (!this.config.email) {
      console.warn("User email not set!, please add email by executing\nstrack cf email ".red + "your@email.here".red);
    }
    (_base2 = this.config).log || (_base2.log = "short");
    (_base3 = this.config).secret || (_base3.secret = createId(this.config.user));
    (_base4 = this.config).defaultState || (_base4.defaultState = "todo");
    (_base5 = this.config).showDone || (_base5.showDone = "false");
    (_base6 = this.config).eof || (_base6.eof = "..");
    (_base7 = this.config).verbose || (_base7.verbose = "true");
    (_base8 = this.config).maxlinesAfterState || (_base8.maxlinesAfterState = "3");
    (_base9 = this.config).sortOrder || (_base9.sortOrder = "asc");
    (_base10 = this.config).showLineNumbers || (_base10.showLineNumbers = "true");
    (_base11 = this.config).trackerFile || (_base11.trackerFile = ".strack");
    (_base12 = this.config).askBeforeCreate || (_base12.askBeforeCreate = "yes");
    return (_base13 = this.config).useZebra || (_base13.useZebra = "yes");
  };
  /*
  Update config with params

  @param {Object} params New config params
  @api public
  */
  Config.prototype.update = function(params) {
    var k, v;
    if (params == null) {
      params = {};
    }
    for (k in params) {
      v = params[k];
      this.config[k] = v;
    }
    return fs.writeFileSync(this.configFile, JSON.stringify(this.config));
  };
  /*
  Dump all config to console

  @api public
  */
  Config.prototype.dump = function() {
    var k, v, _ref, _results;
    console.log("Config parameters:");
    _ref = this.config;
    _results = [];
    for (k in _ref) {
      v = _ref[k];
      _results.push(console.log("" + k + " = " + v));
    }
    return _results;
  };
  /*
  Get config key

  @param {String} key Config key
  @api public
  */
  Config.prototype.get = function(key) {
    return this.config[key];
  };
  /*
  Set config key

  @param {String} key Config key
  @param {String} value Value for key
  @api public
  */
  Config.prototype.set = function(key, value) {
    return this.config[key] = value;
  };
  /*
  Make user dictionary from object

  @return {Object} userDict Dictionary, containing only keys "email" and "user"
  @api public
  */
  Config.prototype.makeUserDict = function() {
    return {
      user: this.config.user,
      email: this.config.email
    };
  };
  return Config;
})();
exports.Config = Config;
/*
Parse text and return it tags and comments

*/
exports.parseText = parseText = function(text) {
  var r, word, _i, _len, _ref;
  r = {
    states: [],
    tags: []
  };
  _ref = text.split(" ");
  for (_i = 0, _len = _ref.length; _i < _len; _i++) {
    word = _ref[_i];
    if (0 === word.indexOf(statePrefix)) {
      r.states.push(word.substring(1));
    } else if (0 === word.indexOf(tagPrefix)) {
      r.tags.push(word.substring(1));
    }
  }
  return r;
};
/*
Get state from text.

@param {String} text Text for search state
@param {String} config Config object
@return {String} state If text have no states, return defaultState
@api public
*/
exports.getState = function(text, config) {
  var word, _i, _len, _ref;
  _ref = text.split(" ");
  for (_i = 0, _len = _ref.length; _i < _len; _i++) {
    word = _ref[_i];
    if (0 === word.indexOf(statePrefix)) {
      return word.substring(1);
    }
  }
  return config.get("defaultState");
};
/*
Replace state from one to another

@param {String} text Text with one state
@param {String} newState New state value
*/
exports.replaceState = function(text, newState) {
  var r, word, _i, _len, _ref;
  r = [];
  _ref = text.split(" ");
  for (_i = 0, _len = _ref.length; _i < _len; _i++) {
    word = _ref[_i];
    if (0 === word.indexOf(statePrefix)) {
      r.push(newState);
    } else {
      r.push(word);
    }
  }
  return r.join(" ");
};
/*
Get word if it matching regexp

@param {String} word Analized word
@param {RegExp} re Regular expression for match
@param {Number} prefixLen Prefix length (for get substring of matching text)
@return {Array|null} result If word match re, return Array of matching part and rest of word.
                            Else, if word not match re, function returns null
@api private
*/
getWord = function(word, re, prefixLen) {
  var m, m2, word2;
  if (prefixLen == null) {
    prefixLen = 1;
  }
  m = word.match(re);
  if (m) {
    word2 = word.substring(prefixLen);
    m2 = word2.match(separatorCharsRe);
    if (m2) {
      return [m2[1], word2.substring(m2[0].length)];
    }
  }
  return null;
};
/*
Push word to result and if makeGrey flag is true, word become grey

@param {Array} result Accumulator array for words
@param {String} uninflectedWord Word that pad to second word to the left
@param {String} word Word string
@param {Boolean} makeGrey Flag, that points to change word color
@api private
*/
pushWords = function(result, uninflectedWord, word, makeGrey) {
  return result.push(uninflectedWord + (makeGrey ? word.grey : word));
};
/*
Apply color or style on text

@param {String} text Text to apply color
@param {String|Array} color Color can be a string or a list of strings
@return {String} coloredText Text with applyed color
@api private
*/
applyColor = function(text, color) {
  var c, _i, _len;
  switch (color) {
    case "black":
      return text.black;
    case "white":
      return text.white;
    case "magenta":
      return text.magenta;
    case "blue":
      return text.blue;
    case "green":
      return text.green;
    case "grey":
      return text.grey;
    case "yellow":
      return text.yellow;
    case "red":
      return text.red;
    case "underline":
      return text.underline;
    case "bold":
      return text.bold;
    case "italic":
      return text.italic;
    case "inverse":
      return text.inverse;
    default:
      if ("object" === typeof color) {
        for (_i = 0, _len = color.length; _i < _len; _i++) {
          c = color[_i];
          text = applyColor(text, c);
        }
        c;
      }
      return text;
  }
};
/*
Colorize string.

@param {String} str Source string
@param {Boolean} flag Color flag
@param {String|Array} trueColor Color(s), that applied if flag is set to true
@param {String|Array} falseColor Color(s), that applied if flag is set to false
@return {String} result Resulting string
@api public
*/
exports.colorizeString = function(str, flag, trueColor, falseColor) {
  if (trueColor == null) {
    trueColor = "green";
  }
  if (falseColor == null) {
    falseColor = "grey";
  }
  return applyColor(str, flag ? trueColor : falseColor);
};
/*
Colorize text for output to terminal.

@param {String} text Text to colorize
@param {String} pattern Search pattern, default - null
@param {Boolean} done Flag, that signal that text must be grey (if done = true)
@return {String} Colored string
@api public
*/
exports.colorizeText = function(text, matchingPattern, done) {
  var result, uninflectedWord, word, words, wrd, _i, _len, _ref;
  if (matchingPattern == null) {
    matchingPattern = null;
  }
  if (done == null) {
    done = false;
  }
  result = [];
  _ref = text.split(" ");
  for (_i = 0, _len = _ref.length; _i < _len; _i++) {
    word = _ref[_i];
    uninflectedWord = "";
    if (0 === word.indexOf(statePrefix)) {
      words = getWord(word, stateRe, statePrefix.length);
      if (words) {
        uninflectedWord = words[0].bold.underline.green;
        if (done) {
          uninflectedWord = uninflectedWord.inverse;
        }
        word = words[1];
      }
    } else if (0 === word.indexOf(tagPrefix)) {
      words = getWord(word, tagRe, tagPrefix.length);
      if (words) {
        uninflectedWord = words[0].underline.magenta;
        word = words[1];
      }
    } else if (0 <= word.indexOf(matchingPattern)) {
      wrd = word.substring(0, matchingPattern.length);
      wrd = wrd.bold.red;
      uninflectedWord = wrd;
      word = word.substring(matchingPattern.length);
    }
    pushWords(result, uninflectedWord, word, done);
  }
  return result.join(" ");
};
/*
Repeat string N times

@param {String} str Source string
@param {Number} N Times to repeat string
@return {String} result string
@api private
*/
times = function(str, N) {
  var i, s;
  s = "";
  for (i = 1; (1 <= N ? i <= N : i >= N); (1 <= N ? i += 1 : i -= 1)) {
    s += str;
  }
  return s;
};
/*
Cut first line of text. If text contains "\n" all chars after
this symbol will be removed from resulting string

@param {String} text Text string.
@param {Number} maxChars Maximum string length in chars, default - 80

@return {String} result Result string. if string length less than maxChars, it will
                        be cuncatenate with spaces
@api public
*/
exports.cutFirstLine = function(text, maxChars) {
  var ind;
  if (maxChars == null) {
    maxChars = 80;
  }
  ind = text.indexOf("\n");
  if (0 <= ind) {
    text = text.substring(0, ind);
  }
  if (maxChars < text.length) {
    return text.slice(0, (maxChars + 1) || 9e9);
  } else {
    return text + times(" ", text.length - maxChars);
  }
};
/*
Format time from datetime string

@param {String} dt Datetime string
@return {String} result Formatted time (hh/mm)
@api public
*/
exports.formatTime = function(dt) {
  var hr, m;
  if ("string" === typeof dt) {
    dt = new Date(Date.parse(dt));
  }
  hr = dt.getHours();
  if (10 > hr) {
    hr = "0" + hr;
  }
  m = dt.getMinutes();
  if (10 > m) {
    m = "0" + m;
  }
  return "" + hr + ":" + m;
};
/*
Format date and time from datetime string

@param {String} dt Datetime string
@return {String} result Formatted time (hh/mm)
@api public
*/
exports.formatDateTime = function(dt) {
  var dd, m;
  if ("string" === typeof dt) {
    dt = new Date(Date.parse(dt));
  }
  dd = dt.toString().split(' ');
  m = dt.getMonth() + 1;
  if (10 > m) {
    m = "0" + m;
  }
  return dd[4] + " " + dd[2] + "." + m + "." + dd[3];
};
/*
List directory, and return list of files, matching mask

@param {String} path Search path
@param {ReqExp} mask Mask for matching file
@param {Boolean} recursive If this flag is set, files will be searched recursively
@return {Array} result List of file pathes
@api public
*/
exports.listDir = listDir = function(path, mask, recursive) {
  var f, file, files, _i, _len, _ref;
  if (mask == null) {
    mask = ".js";
  }
  if (recursive == null) {
    recursive = true;
  }
  files = [];
  _ref = fs.readdirSync(path);
  for (_i = 0, _len = _ref.length; _i < _len; _i++) {
    f = _ref[_i];
    file = path + ("/" + f);
    try {
      if (mask === p.extname(file)) {
        files.push(file);
      } else if (recursive && fs.statSync(file).isDirectory()) {
        listDir(file, mask, recursive).forEach(function(ff) {
          return files.push(ff);
        });
      }
    } catch (err) {

    }
  }
  return files;
};
/*
Read text from standart input and call fn after user input terminator

@param {String} terminator Terminator string
@param {Function} fn Callback function, that accept one parameter - readed data
@api public
*/
exports.readText = function(terminator, fn) {
  var buf, repl, stdin, stdout;
  buf = "";
  stdin = process.openStdin();
  stdout = process.stdout;
  console.log("After entering text write " + terminator.red + " on new line.");
  repl = readline.createInterface(stdin, stdout);
  repl.setPrompt('-> ');
  repl.on('close', function() {
    return stdin.destroy();
  });
  repl.on('line', function(buffer) {
    buffer = buffer.toString();
    if (terminator === buffer) {
      fn(buf);
      process.exit(0);
    }
    buf += buffer + "\n";
    repl.setPrompt("   ");
    return repl.prompt();
  });
  return repl.prompt();
};
/*
Read one line and make call of fn with this line

@param {String} prompt Prompt, default "->"
@param {Function} fn Callback function
*/
exports.readOneLine = function(prompt, fn) {
  var repl, stdin, stdout;
  if (prompt == null) {
    prompt = "->";
  }
  stdin = process.openStdin();
  stdout = process.stdout;
  repl = readline.createInterface(stdin, stdout);
  repl.setPrompt(prompt);
  repl.on('close', function() {
    return stdin.destroy();
  });
  repl.on('line', function(buffer) {
    fn(buffer.toString());
    return process.exit(0);
  });
  return repl.prompt();
};
/*
Edit lines in console

@param {String} text Multiline text add to readline history and can access via keyboard
@param {String} terminator Terminator string
@param {Function} fn Callback function, that called after all text collected
@param {Boolean} lineNum Add line numbers if flag is set, default - true
*/
exports.editTextLines = function(text, terminator, fn, lineNum) {
  var buf, i, l, repl, stdin, stdout, _i, _len, _ref;
  if (lineNum == null) {
    lineNum = true;
  }
  buf = "";
  stdin = process.openStdin();
  stdout = process.stdout;
  console.log("Press ↑ or ↓ to access ticket content");
  console.log("After entering text write " + terminator.red + " on new line.");
  repl = readline.createInterface(stdin, stdout);
  repl.history = [];
  i = 0;
  _ref = text.split("\n");
  for (_i = 0, _len = _ref.length; _i < _len; _i++) {
    l = _ref[_i];
    i++;
    if (lineNum) {
      l += " [" + i + "]";
    }
    repl.history.push(l);
  }
  repl.setPrompt('-> ');
  repl.on('close', function() {
    return stdin.destroy();
  });
  repl.on('line', function(buffer) {
    buffer = buffer.toString();
    if (terminator === buffer) {
      fn(buf);
      process.exit(0);
    }
    buf += buffer + "\n";
    repl.setPrompt("   ");
    return repl.prompt();
  });
  return repl.prompt();
};
/*
Extract from path filename and several directotires above

@param {String} path Full path to file
@return {String} filename Filename with some parent diractories
@api public
*/
exports.extractPartialPath = function(path) {
  return p.basename(p.dirname(path)) + '/' + p.basename(path);
};
/*
Colorize comment number for detail info

@param {Number} num Ordered comment number
@param {String} id Comment id
@param {Number} len Max length of comment id, default - 12
@return {String} result Formated string with comment number and id
@api public
*/
exports.colorizeCommentNumber = function(num, id, len) {
  var n;
  if (len == null) {
    len = 12;
  }
  n = num;
  num = num < 36 ? " ^" + num + " " : "    ";
  if (0 === n % 2) {
    num = num.green.inverse;
  }
  id = id.substring(0, len).yellow;
  return "" + num + " " + id;
};
/*
Make line longer. Increase line length with spaces to len.
If line longer, than length, do nothing

@param {String} line Source line
@param {Number} minLength Minimum line length
@param {String[1]} chr Space character
@return {String} newLine
@api public
*/
exports.makeLineLonger = function(line, len, chr) {
  if (len == null) {
    len = 80;
  }
  if (chr == null) {
    chr = " ";
  }
  while (len > line.length) {
    line += chr;
  }
  return line;
};
/*
Search tags in text and replace them

@param {String} text Text with tags
@param {Function} replaceFn Replace function, call on every tag and must return new string
                            that replace old tagstring
@param {RegExp} tagsRe Regular Expression for searching tags. Default bings to "+tag"
@return {Array} result Result consists of new text at 0 position and tags list at 1 pos
@api public
*/
exports.searchAndReplaceTags = function(text, replaceFn, tagsRe) {
  var match, re, result, t, tags, _i, _len;
  if (tagsRe == null) {
    tagsRe = /(\+\S+)/g;
  }
  tags = [];
  result = text;
  match = text.match(tagsRe);
  if (match) {
    for (_i = 0, _len = match.length; _i < _len; _i++) {
      t = match[_i];
      re = new RegExp(t.replace(/\+/g, "\\+"), "g");
      result = result.replace(re, replaceFn(t.substring(1)));
      tags.push(t.substring(1));
    }
  }
  return [result, tags];
};